from pypact.pact import Pact
import time

pact = Pact()


def check_auth(cmd, device_info):
    pub_key = cmd.get('pubKey')
    sig = cmd.get('sig')
    device_exec = cmd.get('device_exec')
    if pub_key and sig and device_exec:
        verify = pact.crypto.verify(device_exec, pub_key, sig)
        if verify:
            device = device_info
            if device.keys() and pub_key in device['guard']['keys']:
                return True
            else:
                return False
        else:
            return False
    else:
        print("Mission anyone of these variable pubKey, sig, device_exec")
        return False


def validate_expiry(msg):
    expiry_time = msg.get('expiry_time')
    if expiry_time:
        now = time.time().__round__()
        if now < expiry_time:
            return True
        else:
            print("Time expired")
            return False
    else:
        print("expiry_time required")
        return False



import json
import time
import requests
from data_shipper import config
from pypact.pact import Pact

pact = Pact()


def is_cnx_active():
    try:
        requests.head("http://cyberfly.io", timeout=5)
        return True
    except requests.ConnectionError:
        return False


def default_meta(sender="not real"):
    return pact.lang.mk_meta(sender, config.chain_id, 0.000001, 80000, time.time().__round__()-15, 28800)


def get_api_host(network_id):
    if network_id == "testnet04":
        return "https://api.testnet.chainweb.com/chainweb/0.0/testnet04/chain/{}/pact".format(config.chain_id)
    else:
        return "https://api.chainweb.com/chainweb/0.0/mainnet01/chain/{}/pact".format(config.chain_id)


def make_rule(rule: dict) -> str:
    rule = json.loads(rule)
    variable = rule['variable']
    operator = rule['operator']
    value = rule['value']
    if not is_number(value):
        value = '"{}"'.format(value)
    return variable+' '+operator+' '+value


def publish(client, data, key_pair):
    data = json.loads(data)
    device_list = make_list(data['to_devices'])
    for device_id in device_list:
        cmd = make_cmd(data['data'], key_pair)
        try:
            mqtt_publish(client, device_id, cmd)
            print("published to device {}".format(device_id))
        except Exception as e:
            print(e.__str__())


def mqtt_publish(client, topic, cmd):
    payload = json.dumps(cmd)
    client.publish(topic, payload)


def make_cmd(data, key_pair):
    data.update({"expiry_time": time.time().__round__() + 10})
    signed = pact.crypto.sign(json.dumps(data), key_pair)
    signed.update({"device_exec": json.dumps(data)})
    return signed


def is_number(s):
    try:
        complex(s)
    except ValueError:
        return False
    return True


def make_list(s):
    if isinstance(s, list):
        return s
    return [s]



from pypact.pact import Pact
from data_shipper import config, utils
import time


def get_rules(device_id: str, network_id: str, key_pair: dict) -> list:
    pact = Pact()
    pact_code = '({}.{}.read-device-rules "{}")'.format(config.namespace, config.module, device_id)
    cmd = {
        "pactCode": pact_code,
        "envData": {},
        "meta": utils.default_meta(),
        "networkId": network_id,
        "nonce": time.time().__round__() - 15,
        "keyPairs": [key_pair]
    }
    try:
        rules = pact.fetch.local(cmd, utils.get_api_host(network_id))
        if isinstance(rules, dict) and rules.get('result')['status'] == "success":
            return rules.get('result')['data']
        else:
            return []
    except Exception as e:
        print(e.__str__())
        return []


def get_device(device_id: str, network_id: str, key_pair: dict) -> dict:
    while True:
        if utils.is_cnx_active() is True:
            pact = Pact()
            pact_code = '({}.{}.get-device "{}")'.format(config.namespace, config.module, device_id)
            cmd = {
                "pactCode": pact_code,
                "envData": {},
                "meta": utils.default_meta(),
                "networkId": network_id,
                "nonce": time.time().__round__() - 15,
                "keyPairs": [key_pair]
            }
            try:
                devices = pact.fetch.local(cmd, utils.get_api_host(network_id))
                if isinstance(devices, dict) and devices.get('result')['status'] == "success":
                    return devices.get('result')['data']
                else:
                    return {}
            except Exception as e:
                print(e.__str__())
                return {}
            break
        else:
            pass






import json

from typing import Callable
import paho.mqtt.client as mqtt
import rule_engine
from data_shipper import config, auth, api, utils

mqttc = mqtt.Client(clean_session=True)


class CyberflyDataShipper:
    def __init__(self, device_id: str, key_pair: dict, network_id: str = "mainnet01"):
        self.key_pair = key_pair
        self.network_id = network_id
        self.device_data = {}
        self.device_id = device_id
        self.account = "k:" + self.key_pair.get("publicKey")
        self.caller = default_caller
        self.mqtt_client = mqttc
        self.topic = device_id
        self.mqtt_client.user_data_set(self)
        self.mqtt_client.on_connect = on_connect
        self.mqtt_client.on_message = on_received
        self.rules = []
        self.device_info = {}
        self.update_device()
        self.update_rules()
        self.run(config.mqtt_broker, config.mqtt_port)

    def update_data(self, key: str, value):
        self.device_data.update({key: value})

    def on_message(self) -> Callable:
        def decorator(callback_function):
            self.caller = callback_function

        return decorator

    def run(self, host: str, port: int) -> None:
        print("trying to connect")
        try:
            self.mqtt_client.connect(host, port, 60)
        except Exception as e:
            print(e.__str__())
        try:
            self.mqtt_client.loop_start()
        except Exception as e:
            print(e.__str__())

    def process_data(self, data: dict):
        rules = self.rules
        if len(rules) == 0:
            self.update_rules()
        context = rule_engine.Context(default_value=None)
        for rule in rules:
            rul = rule_engine.Rule(utils.make_rule(rule['rule']), context=context)
            try:
                if rul.matches(data):
                    utils.publish(self.mqtt_client, rule['action'], self.key_pair)
            except Exception as e:
                print(e.__str__())

    def publish(self, topic, msg):
        signed = utils.make_cmd(msg, self.key_pair)
        utils.mqtt_publish(self.mqtt_client, topic, signed)

    def update_rules(self):
        rules = api.get_rules(self.device_id, self.network_id, self.key_pair)
        self.rules = rules

    def update_device(self):
        device = api.get_device(self.device_id, self.network_id, self.key_pair)
        self.device_info = device


def on_connect(client: mqtt.Client, mqtt_class: CyberflyDataShipper, __flags, received_code: int) -> None:
    print("Connected with result code " + str(received_code))
    client.subscribe(mqtt_class.topic)


def on_received(__client: mqtt.Client, mqtt_class: CyberflyDataShipper, msg: mqtt.MQTTMessage) -> None:
    json_string = msg.payload.decode("utf-8")
    try:
        json_data = json.loads(json_string)
        device_exec = json.loads(json_data['device_exec'])
        response_topic = device_exec.get('mqtt_response_topic')
        if auth.validate_expiry(device_exec) \
                and auth.check_auth(json_data, mqtt_class.device_info):
            try:
                if device_exec.get('update_rules'):
                    mqtt_class.update_rules()
                if device_exec.get('update_device'):
                    mqtt_class.update_device()
                mqtt_class.caller(device_exec)
                if response_topic:
                    signed = utils.make_cmd({"info": "success"}, mqtt_class.key_pair)
                    utils.mqtt_publish(__client, response_topic, signed)
            except Exception as e:
                signed = utils.make_cmd({"info": "error"}, mqtt_class.key_pair)
                utils.mqtt_publish(__client, response_topic, signed)
                print(e.__str__())
        else:
            print(mqtt_class.device_info)
            print("auth failed")
    except Exception as e:
        print(e.__str__())


def default_caller(data):
    pass
